/****************************************************************
BeebEm - BBC Micro and Master 128 Emulator
Copyright (C) 1994  David Alan Gilbert
Copyright (C) 1997  Mike Wyatt
Copyright (C) 2001  Richard Gellman

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
****************************************************************/

/* 6502 core - 6502 emulator core - David Alan Gilbert 16/10/94 */
/* Mike Wyatt 7/6/97 - Added undocumented instructions */
/* Copied for 65C02 Tube core - 13/04/01 */

#include <iostream>
#include <fstream>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "6502core.h"
#include "log.h"
#include "main.h"
#include "beebmem.h"
#include "tube.h"
#include "debug.h"
#include "uefstate.h"
#include "z80mem.h"
#include "z80.h"
#include "Arm.h"
#include "sprowcopro.h"

#ifdef WIN32
#include <windows.h>
#define INLINE inline
#else
#define INLINE
#endif

// Some interrupt set macros
#define SETTUBEINT(a) TubeintStatus |= 1 << a
#define RESETTUBEINT(a) TubeintStatus &= ~(1 << a)

static int CurrentInstruction;
unsigned char TubeRam[65536];
Tube TubeType;

CycleCountT TotalTubeCycles = 0;

unsigned char old_readHIOAddr = 0;
unsigned char old_readHTmpData = 0;

unsigned char old_readPIOAddr = 0;
unsigned char old_readPTmpData = 0;

unsigned char old_writeHIOAddr = 0;
unsigned char old_writeHTmpData = 0;

unsigned char old_writePIOAddr = 0;
unsigned char old_writePTmpData = 0;

int TubeProgramCounter;
static int PreTPC; // Previous Tube Program Counter;
static int Accumulator, XReg, YReg;
static unsigned char StackReg, PSR;
static unsigned char IRQCycles;

unsigned char TubeintStatus = 0; // bit set (nums in IRQ_Nums) if interrupt being caused
unsigned char TubeNMIStatus = 0; // bit set (nums in NMI_Nums) if NMI being caused
static bool TubeNMILock = false; // Well I think NMI's are maskable - to stop repeated NMI's - the lock is released when an RTI is done

typedef int int16;

/* Note how GETCFLAG is special since being bit 0 we don't need to test it to get a clean 0/1 */
#define GETCFLAG ((PSR & FlagC))
#define GETZFLAG ((PSR & FlagZ)>0)
#define GETIFLAG ((PSR & FlagI)>0)
#define GETDFLAG ((PSR & FlagD)>0)
#define GETBFLAG ((PSR & FlagB)>0)
#define GETVFLAG ((PSR & FlagV)>0)
#define GETNFLAG ((PSR & FlagN)>0)

static const int TubeCyclesTable[] = {
/*0 1 2 3 4 5 6 7 8 9 a b c d e f */
  7,6,2,1,5,3,5,5,3,2,2,1,6,4,6,5, /* 0 */
  2,5,5,1,5,4,6,5,2,4,2,1,6,4,6,5, /* 1 */
  6,6,2,1,3,3,5,5,4,2,2,1,4,4,6,5, /* 2 */
  2,5,5,1,4,4,6,5,2,4,2,1,4,4,6,5, /* 3 */
  6,6,2,1,3,3,5,5,3,2,2,1,3,4,6,5, /* 4 */
  2,5,5,1,4,4,6,5,2,4,3,1,8,4,6,5, /* 5 */
  6,6,2,1,3,3,5,5,4,2,2,1,6,4,6,5, /* 6 */
  2,5,5,1,4,4,6,5,2,4,4,1,6,4,6,5, /* 7 */
  3,6,2,1,3,3,3,5,2,2,2,1,4,4,4,5, /* 8 */
  2,6,5,1,4,4,4,5,2,5,2,1,4,5,5,5, /* 9 */
  2,6,2,1,3,3,3,5,2,2,2,1,4,4,4,5, /* a */
  2,5,5,1,4,4,4,5,2,4,2,1,4,4,4,5, /* b */
  2,6,2,1,3,3,5,5,2,2,2,1,4,4,6,5, /* c */
  2,5,5,1,4,4,6,5,2,4,3,1,4,4,7,5, /* d */
  2,6,2,1,3,3,5,5,2,2,2,1,4,4,6,5, /* e */
  2,5,5,1,4,4,6,5,2,4,4,1,4,4,7,5  /* f */
};

/* The number of TubeCycles to be used by the current instruction - exported to
   allow fernangling by memory subsystem */
unsigned int TubeCycles;

static bool Branched; // true if the instruction branched

/* A macro to speed up writes - uses a local variable called 'tmpaddr' */
#define TUBEREADMEM_FAST(a) ((a<0xfef8)?TubeRam[a]:TubeReadMem(a))
#define TUBEWRITEMEM_FAST(Address, Value) if (Address<0xfef8) TubeRam[Address]=Value; else TubeWriteMem(Address,Value);
#define TUBEWRITEMEM_DIRECT(Address, Value) TubeRam[Address]=Value;
#define TUBEFASTWRITE(addr,val) tmpaddr=addr; if (tmpaddr<0xfef8) TUBEWRITEMEM_DIRECT(tmpaddr,val) else TubeWriteMem(tmpaddr,val);

// Local fns
static void Reset65C02();

// Staus bits
enum TubeFlags {
	TubeQ=1,         // Host IRQ from reg 4
	TubeI=2,         // Parasite IRQ from reg 1
	TubeJ=4,         // Parasite IRQ from reg 4
	TubeM=8,         // Parasite NMI from reg 3
	TubeV=16,        // Two byte op for reg 3
	TubeP=32,        // Parasite processor reset
	TubeT=64,        // Tube reset (write only)
	TubeNotFull=64,  // Reg not full (read only)
	TubeS=128,       // Set control flags mask (write only)
	TubeDataAv=128   // Data available (read only)
};

// Tube registers
unsigned char R1Status; // Q,I,J,M,V,P flags

unsigned char R1PHData[TubeBufferLength * 2];
int R1PHPtr;
unsigned char R1HStatus;
unsigned char R1HPData;
unsigned char R1PStatus;

unsigned char R2PHData;
unsigned char R2HStatus;
unsigned char R2HPData;
unsigned char R2PStatus;

unsigned char R3PHData[2];
unsigned char R3PHPtr;
unsigned char R3HStatus;
unsigned char R3HPData[2];
unsigned char R3HPPtr;
unsigned char R3PStatus;

unsigned char R4PHData;
unsigned char R4HStatus;
unsigned char R4HPData;
unsigned char R4PStatus;

/*-------------------------------------------------------------------*/
// Tube interupt functions
void UpdateR1Interrupt(void) {
	if ((R1Status & TubeI) && (R1PStatus & TubeDataAv))
		SETTUBEINT(R1);
	else
		RESETTUBEINT(R1);
}

void UpdateR4Interrupt(void) {
	if ((R1Status & TubeJ) && (R4PStatus & TubeDataAv))
		SETTUBEINT(R4);
	else
		RESETTUBEINT(R4);
}

void UpdateR3Interrupt(void) {
	if ((R1Status & TubeM) && !(R1Status & TubeV) &&
		( (R3HPPtr > 0) || (R3PHPtr == 0) ))
		TubeNMIStatus|=(1<<R3);
	else if ((R1Status & TubeM) && (R1Status & TubeV) &&
		( (R3HPPtr > 1) || (R3PHPtr == 0) ))
		TubeNMIStatus|=(1<<R3);
	else
		TubeNMIStatus&=~(1<<R3);
}

void UpdateHostR4Interrupt(void) {
	if ((R1Status & TubeQ) && (R4HStatus & TubeDataAv))
		intStatus|=(1<<tube);
	else
		intStatus&=~(1<<tube);
}

/*-------------------------------------------------------------------*/
// Torch tube memory/io handling functions

static bool TorchTubeActive = false;

void UpdateInterrupts()
{
	UpdateR1Interrupt();
	UpdateR3Interrupt();
	UpdateR4Interrupt();
	UpdateHostR4Interrupt();
}

unsigned char ReadTorchTubeFromHostSide(unsigned char IOAddr)
{
	unsigned char TmpData = 0xff;

	if (!TorchTubeActive)
		return MachineType == Model::Master128 ? 0xff : 0xfe;

	switch (IOAddr) {
	case 0:
		TmpData=R1HStatus | R1Status;
		break;

	case 1:
		TmpData=R1PHData[0];
		R1HStatus&=~TubeDataAv;
		R1PStatus|=TubeNotFull;
		break;

// Data available in Z80 tube ?
// Bit #2 says data available to read in 0xfee1
// Bit #10 says room available to write to 0xfee1

	case 0x0d:
		TmpData = 0;
		if (R1HStatus & TubeDataAv) TmpData |= 0x02;
		if (R1HStatus & TubeNotFull) TmpData |= 0x10;
		break;

	case 0x10:
		// trace = 1;
		break;
	}

	// WriteLog("ReadTorchTubeFromHostSide - Addr = %02x, Value = %02x\n", (int)IOAddr, (int)TmpData);

	if (DebugEnabled) {
		char info[200];
		sprintf(info, "Tube: Read from host, addr %X value %02X", (int)IOAddr, (int)TmpData);
		DebugDisplayTrace(DebugType::Tube, true, info);
	}

	return TmpData;
}

void WriteTorchTubeFromHostSide(unsigned char IOAddr,unsigned char IOData)
{
	// WriteLog("WriteTorchTubeFromHostSide - Addr = %02x, Value = %02x\n", (int)IOAddr, (int)IOData);

	if (DebugEnabled) {
		char info[200];
		sprintf(info, "Tube: Write from host, addr %X value %02X", (int)IOAddr, (int)IOData);
		DebugDisplayTrace(DebugType::Tube, true, info);
	}

	if (IOAddr == 0x02 && IOData == 0xff)
	{
		TorchTubeActive = true;
	}

	switch (IOAddr) {
	case 1:
		// S bit controls write of status flags
		if (IOData & TubeS)
			R1Status|=(IOData & 0x3f);
		else
			R1Status&=~IOData;

		// Reset required?
		if (R1Status & TubeP)
			Reset65C02();
		if (R1Status & TubeT)
			ResetTube();

		// Update interrupt flags
		UpdateR1Interrupt();
		UpdateR3Interrupt();
		UpdateR4Interrupt();
		UpdateHostR4Interrupt();
		break;
	case 0:
		R1HPData = IOData;
		R1PStatus|=TubeDataAv;
		R1HStatus&=~TubeNotFull;
		UpdateR1Interrupt();
		break;

// Echo back to tube ?

	case 0x08 :
		// WriteTorchTubeFromHostSide(1, IOData);
		break;

	case 0x0c :
		if (IOData == 0xaa)
		{
			init_z80();
		}
		break;

	case 0x0e:
		break;
	}
}

unsigned char ReadTorchTubeFromParasiteSide(unsigned char IOAddr)
{
	unsigned char TmpData;

	switch (IOAddr) {
	case 0:
		TmpData=R1PStatus | R1Status;
		break;
	case 1:
		TmpData = R1HPData;
		R1PStatus&=~TubeDataAv;
		R1HStatus|=TubeNotFull;
		UpdateR1Interrupt();
		break;
	}

	// WriteLog("ReadTorchTubeFromParasiteSide - Addr = %02x, Value = %02x\n", (int)IOAddr, (int)TmpData);

	if (DebugEnabled) {
		char info[200];
		sprintf(info, "Tube: Read from para, addr %X value %02X", (int)IOAddr, (int)TmpData);
		DebugDisplayTrace(DebugType::Tube, false, info);
	}

	return TmpData;
}

void WriteTorchTubeFromParasiteSide(unsigned char IOAddr,unsigned char IOData)
{
	// WriteLog("WriteTorchTubeFromParasiteSide - Addr = %02x, Value = %02x\n", (int)IOAddr, (int)IOData);

	if (DebugEnabled) {
		char info[200];
		sprintf(info, "Tube: Write from para, addr %X value %02X", (int)IOAddr, (int)IOData);
		DebugDisplayTrace(DebugType::Tube, false, info);
	}

	switch (IOAddr) {
	case 1:
		R1PHData[0]=IOData;
		R1HStatus|=TubeDataAv;
		R1PStatus&=~TubeNotFull;
		break;
	}
}

/*-------------------------------------------------------------------*/
// Tube memory/io handling functions

unsigned char ReadTubeFromHostSide(unsigned char IOAddr) {
	unsigned char TmpData,TmpCntr;

	if (TubeType == Tube::None)
		return MachineType == Model::Master128 ? 0xff : 0xfe;

	switch (IOAddr) {
	case 0:
		TmpData=R1HStatus | R1Status;
		break;
	case 1:
		TmpData=R1PHData[0];
		if (R1PHPtr>0) {
			for (TmpCntr=1;TmpCntr<TubeBufferLength;TmpCntr++)
				R1PHData[TmpCntr-1]=R1PHData[TmpCntr]; // Shift FIFO Buffer
			R1PHPtr--; // Shift FIFO Pointer
			if (R1PHPtr == 0)
				R1HStatus&=~TubeDataAv;
			R1PStatus|=TubeNotFull;
		}
		break;
	case 2:
		TmpData=R2HStatus;
		break;
	case 3:
		TmpData=R2PHData;
		if (R2HStatus & TubeDataAv) {
			R2HStatus&=~TubeDataAv;
			R2PStatus|=TubeNotFull;
		}
		break;
	case 4:
		TmpData=R3HStatus;
		break;
	case 5:
		TmpData=R3PHData[0];
		if (R3PHPtr>0) {
			R3PHData[0]=R3PHData[1]; // Shift FIFO Buffer
			R3PHPtr--; // Shift FIFO Pointer
			if (R3PHPtr == 0) {
				R3HStatus&=~TubeDataAv;
				R3PStatus|=TubeNotFull;
			}
		}
		UpdateR3Interrupt();
		break;
	case 6:
		TmpData=R4HStatus;
		break;
	case 7:
		TmpData=R4PHData;
		if (R4HStatus & TubeDataAv) {
			R4HStatus&=~TubeDataAv;
			R4PStatus|=TubeNotFull;
		}
		UpdateHostR4Interrupt();
		break;
	}

	if (DebugEnabled && (old_readHIOAddr != IOAddr || old_readHTmpData != TmpData)) {
		char info[200];
		sprintf(info, "Tube: Read from host, addr %X value %02X\r\n", (int)IOAddr, (int)TmpData);
		DebugDisplayTrace(DebugType::Tube, true, info);
	}

	old_readHTmpData = TmpData;
	old_readHIOAddr = IOAddr;

	return TmpData;
}

void WriteTubeFromHostSide(unsigned char IOAddr,unsigned char IOData) {
	if (TubeType == Tube::None)
		return;

	if (DebugEnabled) {
		char info[200];
		sprintf(info, "Tube: Write from host, addr %X value %02X", (int)IOAddr, (int)IOData);
		DebugDisplayTrace(DebugType::Tube, true, info);
	}

	switch (IOAddr) {
	case 0:
		// S bit controls write of status flags
		if (IOData & TubeS)
			R1Status|=(IOData & 0x3f);
		else
			R1Status&=~IOData;

		// Reset required?
		if (R1Status & TubeP)
			Reset65C02();
		if (R1Status & TubeT)
			ResetTube();

		// Update interrupt flags
		UpdateR1Interrupt();
		UpdateR3Interrupt();
		UpdateR4Interrupt();
		UpdateHostR4Interrupt();
		break;
	case 1:
		R1HPData=IOData;
		R1PStatus|=TubeDataAv;
		R1HStatus&=~TubeNotFull;
		UpdateR1Interrupt();
		break;
	case 3:
		R2HPData=IOData;
		R2PStatus|=TubeDataAv;
		R2HStatus&=~TubeNotFull;
		break;
	case 5:
		if (R1Status & TubeV) {
			if (R3HPPtr < 2)
				R3HPData[R3HPPtr++]=IOData;
			if (R3HPPtr == 2) {
				R3PStatus|=TubeDataAv;
				R3HStatus&=~TubeNotFull;
			}
		}
		else {
			R3HPPtr=0;
			R3HPData[R3HPPtr++]=IOData;
			R3PStatus|=TubeDataAv;
			R3HStatus&=~TubeNotFull;
		}
		UpdateR3Interrupt();
		break;
	case 7:
		R4HPData=IOData;
		R4PStatus|=TubeDataAv;
		R4HStatus&=~TubeNotFull;
		UpdateR4Interrupt();
		break;
	}

	// UpdateInterrupts();
}

unsigned char ReadTubeFromParasiteSide(unsigned char IOAddr) {
	unsigned char TmpData;

	if (TubeType == Tube::TorchZ80)
		return ReadTorchTubeFromHostSide(IOAddr);

	switch (IOAddr) {
	case 0:
		TmpData=R1PStatus | R1Status;
		break;
	case 1:
		TmpData=R1HPData;
		if (R1PStatus & TubeDataAv) {
			R1PStatus&=~TubeDataAv;
			R1HStatus|=TubeNotFull;
		}
		UpdateR1Interrupt();
		break;
	case 2:
		TmpData=R2PStatus;
		break;
	case 3:
		TmpData=R2HPData;
		if (R2PStatus & TubeDataAv) {
			R2PStatus&=~TubeDataAv;
			R2HStatus|=TubeNotFull;
		}
		break;
	case 4:
		TmpData=R3PStatus;
		// Tube Spec says top bit in R3PStatus has value 'N', which looks like it is
		// the same as the PNMI status (i.e. H->P data available OR P->H not full).
		if (R3PHPtr == 0)
			TmpData |= 128;
		break;
	case 5:
		TmpData=R3HPData[0];
		if (R3HPPtr>0) {
			R3HPData[0]=R3HPData[1]; // Shift FIFO Buffer
			R3HPPtr--; // Shift FIFO Pointer
			if (R3HPPtr == 0) {
				R3PStatus&=~TubeDataAv;
				R3HStatus|=TubeNotFull;
			}
		}
		UpdateR3Interrupt();
		break;
	case 6:
		TmpData=R4PStatus;
		break;
	case 7:
		TmpData=R4HPData;
		if (R4PStatus & TubeDataAv) {
			R4PStatus&=~TubeDataAv;
			R4HStatus|=TubeNotFull;
		}
		UpdateR4Interrupt();
		break;
	}

	// UpdateInterrupts();

	if (DebugEnabled && (old_readPIOAddr != IOAddr || old_readPTmpData != TmpData)) {
		char info[200];
		sprintf(info, "Tube: Read from para, addr %X value %02X\r\n", (int)IOAddr, (int)TmpData);
		DebugDisplayTrace(DebugType::Tube, false, info);
	}

	old_readPTmpData = TmpData;
	old_readPIOAddr = IOAddr;

	// UpdateInterrupts();
	return TmpData;
}

void WriteTubeFromParasiteSide(unsigned char IOAddr,unsigned char IOData)
{
	if (TubeType == Tube::TorchZ80)
	{
		WriteTorchTubeFromParasiteSide(IOAddr, IOData);
		return;
	}

	if (DebugEnabled) {
		char info[200];
		sprintf(info, "Tube: Write from para, addr %X value %02X\r\n", (int)IOAddr, (int)IOData);
		DebugDisplayTrace(DebugType::Tube, false, info);
	}

	switch (IOAddr) {
	case 0:
		// Cannot write status flags from parasite
		break;
	case 1:
		if (R1PHPtr<TubeBufferLength) {
			R1PHData[R1PHPtr++]=IOData;
			R1HStatus|=TubeDataAv;
			if (R1PHPtr==TubeBufferLength)
				R1PStatus&=~TubeNotFull;
		}
		break;
	case 3:
		R2PHData=IOData;
		R2HStatus|=TubeDataAv;
		R2PStatus&=~TubeNotFull;
		break;
	case 5:
		if (R1Status & TubeV) {
			if (R3PHPtr < 2)
				R3PHData[R3PHPtr++]=IOData;
			if (R3PHPtr == 2) {
				R3HStatus|=TubeDataAv;
				R3PStatus&=~TubeNotFull;
			}
		}
		else {
			R3PHPtr=0;
			R3PHData[R3PHPtr++]=IOData;
			R3HStatus|=TubeDataAv;
			R3PStatus&=~TubeNotFull;
		}
		UpdateR3Interrupt();
		break;
	case 7:
		R4PHData=IOData;
		R4HStatus|=TubeDataAv;
		R4PStatus&=~TubeNotFull;
		UpdateHostR4Interrupt();
		break;
	}
    //UpdateInterrupts();
}

/*----------------------------------------------------------------------------*/
void TubeWriteMem(unsigned int IOAddr,unsigned char IOData) {
	if (IOAddr>=0xff00 || IOAddr<0xfef8)
		TubeRam[IOAddr]=IOData;
	else
		WriteTubeFromParasiteSide(IOAddr-0xfef8,IOData);
}

unsigned char TubeReadMem(unsigned int IOAddr) {
	if (IOAddr>=0xff00 || IOAddr<0xfef8)
		return(TubeRam[IOAddr]);
	else
		return(ReadTubeFromParasiteSide(IOAddr-0xfef8));
}

// Get a two byte address from the program counter, and then post inc
// the program counter
#define GETTWOBYTEFROMPC(var) \
	var = TubeRam[TubeProgramCounter++]; \
	var |= (TubeRam[TubeProgramCounter++] << 8);

/*----------------------------------------------------------------------------*/
INLINE void Carried() {
	// Correct cycle count for indirection across page boundary
	if (((CurrentInstruction & 0xf) == 0x1 ||
	     (CurrentInstruction & 0xf) == 0x9 ||
	     (CurrentInstruction & 0xf) == 0xd) &&
	     (CurrentInstruction & 0xf0) != 0x90)
	{
		TubeCycles++;
	}
	else if (CurrentInstruction == 0xBC ||
	         CurrentInstruction == 0xBE)
	{
		TubeCycles++;
	}
}

/*----------------------------------------------------------------------------*/
/* Set the Z flag if 'in' is 0, and N if bit 7 is set - leave all other bits  */
/* untouched.                                                                 */
INLINE static void SetPSRZN(const unsigned char in) {
  PSR&=~(FlagZ | FlagN);
  PSR|=((in==0)<<1) | (in & 128);
}

/*----------------------------------------------------------------------------*/
/* Note: n is 128 for true - not 1                                            */
INLINE static void SetPSR(int mask,int c,int z,int i,int d,int b, int v, int n) {
  PSR&=~mask;
  PSR|=c | (z<<1) | (i<<2) | (d<<3) | (b<<4) | (v<<6) | n;
}

/*----------------------------------------------------------------------------*/
/* NOTE!!!!! n is 128 or 0 - not 1 or 0                                       */
INLINE static void SetPSRCZN(int c,int z, int n) {
  PSR&=~(FlagC | FlagZ | FlagN);
  PSR|=c | (z<<1) | n;
}

/*----------------------------------------------------------------------------*/
INLINE static void Push(unsigned char ToPush) {
  TUBEWRITEMEM_DIRECT(0x100+StackReg,ToPush);
  StackReg--;
} /* Push */

/*----------------------------------------------------------------------------*/
INLINE static unsigned char Pop(void) {
  StackReg++;
  return(TubeRam[0x100+StackReg]);
} /* Pop */

/*----------------------------------------------------------------------------*/
INLINE static void PushWord(int16 topush) {
  Push((topush>>8) & 255);
  Push(topush & 255);
} /* PushWord */

/*----------------------------------------------------------------------------*/
INLINE static int16 PopWord() {
  int16 RetValue;

  RetValue=Pop();
  RetValue|=(Pop()<<8);
  return(RetValue);
} /* PopWord */

/*-------------------------------------------------------------------------*/

// Relative addressing mode handler

INLINE static int16 RelAddrModeHandler_Data() {
	// For branches - is this correct - i.e. is the program counter incremented
	// at the correct time?
	int EffectiveAddress = (signed char)TubeRam[TubeProgramCounter++];
	EffectiveAddress += TubeProgramCounter;

	return EffectiveAddress;
}

/*----------------------------------------------------------------------------*/
INLINE static void ADCInstrHandler(int16 operand) {
  /* NOTE! Not sure about C and V flags */
  if (!GETDFLAG) {
    int TmpResultC = Accumulator + operand + GETCFLAG;
    int TmpResultV = (signed char)Accumulator + (signed char)operand + GETCFLAG;
    Accumulator = TmpResultC & 255;
    SetPSR(FlagC | FlagZ | FlagV | FlagN, (TmpResultC & 256) > 0,
      Accumulator == 0, 0, 0, 0, ((Accumulator & 128) > 0) ^ (TmpResultV < 0),
      Accumulator & 128);
  } else {
    /* Z flag determined from 2's compl result, not BCD result! */
    int TmpResult = Accumulator + operand + GETCFLAG;
    int ZFlag = (TmpResult & 0xff) == 0;

    int ln = (Accumulator & 0xf) + (operand & 0xf) + GETCFLAG;

    int TmpCarry = 0;

    if (ln > 9) {
      ln += 6;
      ln &= 0xf;
      TmpCarry = 0x10;
    }

    int hn = (Accumulator & 0xf0) + (operand & 0xf0) + TmpCarry;
    /* N and V flags are determined before high nibble is adjusted.
       NOTE: V is not always correct */
    int NFlag = hn & 128;
    int VFlag = (hn ^ Accumulator) & 128 && !((Accumulator ^ operand) & 128);

    int CFlag = 0;

    if (hn > 0x90) {
      hn += 0x60;
      hn &= 0xf0;
      CFlag = 1;
    }

    Accumulator = hn | ln;

    ZFlag = Accumulator == 0;
    NFlag = Accumulator & 128;

    SetPSR(FlagC | FlagZ | FlagV | FlagN, CFlag, ZFlag, 0, 0, 0, VFlag, NFlag);
  }
} /* ADCInstrHandler */

/*----------------------------------------------------------------------------*/
INLINE static void ANDInstrHandler(int16 operand) {
  Accumulator=Accumulator & operand;
  PSR&=~(FlagZ | FlagN);
  PSR|=((Accumulator==0)<<1) | (Accumulator & 128);
} /* ANDInstrHandler */

INLINE static void ASLInstrHandler(int16 address) {
  unsigned char oldVal,newVal;
  oldVal=TUBEREADMEM_FAST(address);
  newVal=(((unsigned int)oldVal)<<1) & 254;
  TUBEWRITEMEM_FAST(address,newVal);
  SetPSRCZN((oldVal & 128)>0, newVal==0,newVal & 128);
} /* ASLInstrHandler */

INLINE static void TRBInstrHandler(int16 address) {
	unsigned char oldVal,newVal;
	oldVal=TUBEREADMEM_FAST(address);
	newVal=(Accumulator ^ 255) & oldVal;
    TUBEWRITEMEM_FAST(address,newVal);
    PSR&=253;
	PSR|=((Accumulator & oldVal)==0) ? 2 : 0;
} // TRBInstrHandler

INLINE static void TSBInstrHandler(int16 address) {
	unsigned char oldVal,newVal;
	oldVal=TUBEREADMEM_FAST(address);
	newVal=Accumulator | oldVal;
    TUBEWRITEMEM_FAST(address,newVal);
    PSR&=253;
	PSR|=((Accumulator & oldVal)==0) ? 2 : 0;
} // TSBInstrHandler

INLINE static void ASLInstrHandler_Acc(void) {
  unsigned char oldVal,newVal;
  /* Accumulator */
  oldVal=Accumulator;
  Accumulator=newVal=(((unsigned int)Accumulator)<<1) & 254;
  SetPSRCZN((oldVal & 128)>0, newVal==0,newVal & 128);
} /* ASLInstrHandler_Acc */

INLINE static void BCCInstrHandler(void) {
  if (!GETCFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
} /* BCCInstrHandler */

INLINE static void BCSInstrHandler(void) {
  if (GETCFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
} /* BCSInstrHandler */

INLINE static void BEQInstrHandler(void) {
  if (GETZFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
} /* BEQInstrHandler */

INLINE static void BITInstrHandler(int16 operand) {
  PSR&=~(FlagZ | FlagN | FlagV);
  /* z if result 0, and NV to top bits of operand */
  PSR|=(((Accumulator & operand)==0)<<1) | (operand & 192);
} /* BITInstrHandler */

INLINE static void BITImmedInstrHandler(int16 operand)
{
	PSR &= ~FlagZ;
	// Z if result 0, and NV to top bits of operand
	PSR |= (((Accumulator & operand) == 0)<<1);
}

INLINE static void BMIInstrHandler(void) {
  if (GETNFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
} /* BMIInstrHandler */

INLINE static void BNEInstrHandler(void) {
  if (!GETZFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
} /* BNEInstrHandler */

INLINE static void BPLInstrHandler(void) {
  if (!GETNFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
}

INLINE static void BRKInstrHandler(void) {
  PushWord(TubeProgramCounter+1);
  SetPSR(FlagB,0,0,0,0,1,0,0); /* Set B before pushing */
  Push(PSR);
  SetPSR(FlagI,0,0,1,0,0,0,0); /* Set I after pushing - see Birnbaum */
  TubeProgramCounter=TubeReadMem(0xfffe) | (TubeReadMem(0xffff)<<8);
} /* BRKInstrHandler */

INLINE static void BVCInstrHandler(void) {
  if (!GETVFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
} /* BVCInstrHandler */

INLINE static void BVSInstrHandler(void) {
  if (GETVFLAG) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
  } else TubeProgramCounter++;
} /* BVSInstrHandler */

INLINE static void BRAInstrHandler(void) {
    TubeProgramCounter=RelAddrModeHandler_Data();
    Branched = true;
} /* BRAnstrHandler */

INLINE static void CMPInstrHandler(int16 operand) {
  /* NOTE! Should we consult D flag ? */
  unsigned char result=Accumulator-operand;
  unsigned char CFlag;
  CFlag=0; if (Accumulator>=operand) CFlag=FlagC;
  SetPSRCZN(CFlag, Accumulator==operand,result & 128);
} /* CMPInstrHandler */

INLINE static void CPXInstrHandler(int16 operand) {
  unsigned char result=(XReg-operand);
  SetPSRCZN(XReg>=operand, XReg==operand,result & 128);
} /* CPXInstrHandler */

INLINE static void CPYInstrHandler(int16 operand) {
  unsigned char result=(YReg-operand);
  SetPSRCZN(YReg>=operand, YReg==operand,result & 128);
} /* CPYInstrHandler */

INLINE static void DECInstrHandler(int16 address) {
  unsigned char val;

  val=TUBEREADMEM_FAST(address);

  val=(val-1);

  TUBEWRITEMEM_FAST(address,val);
  SetPSRZN(val);
} /* DECInstrHandler */

INLINE static void DEXInstrHandler(void) {
  XReg=(XReg-1) & 255;
  SetPSRZN(XReg);
} /* DEXInstrHandler */

INLINE static void DEAInstrHandler(void) {
  Accumulator=(Accumulator-1) & 255;
  SetPSRZN(Accumulator);
} /* DEAInstrHandler */

INLINE static void EORInstrHandler(int16 operand) {
  Accumulator^=operand;
  SetPSRZN(Accumulator);
} /* EORInstrHandler */

INLINE static void INCInstrHandler(int16 address) {
  unsigned char val;

  val=TUBEREADMEM_FAST(address);

  val=(val+1) & 255;

  TUBEWRITEMEM_FAST(address,val);
  SetPSRZN(val);
} /* INCInstrHandler */

INLINE static void INXInstrHandler(void) {
  XReg+=1;
  XReg&=255;
  SetPSRZN(XReg);
} /* INXInstrHandler */

INLINE static void INAInstrHandler(void) {
  Accumulator+=1;
  Accumulator&=255;
  SetPSRZN(Accumulator);
} /* INAInstrHandler */

INLINE static void JSRInstrHandler(int16 address) {
  PushWord(TubeProgramCounter-1);
  TubeProgramCounter=address;
} /* JSRInstrHandler */

INLINE static void LDAInstrHandler(int16 operand) {
  Accumulator=operand;
  SetPSRZN(Accumulator);
} /* LDAInstrHandler */

INLINE static void LDXInstrHandler(int16 operand) {
  XReg=operand;
  SetPSRZN(XReg);
} /* LDXInstrHandler */

INLINE static void LDYInstrHandler(int16 operand) {
  YReg=operand;
  SetPSRZN(YReg);
} /* LDYInstrHandler */

INLINE static void LSRInstrHandler(int16 address) {
  unsigned char oldVal,newVal;
  oldVal=TUBEREADMEM_FAST(address);
  newVal=(((unsigned int)oldVal)>>1) & 127;
  TUBEWRITEMEM_FAST(address,newVal);
  SetPSRCZN((oldVal & 1)>0, newVal==0,0);
} /* LSRInstrHandler */

INLINE static void LSRInstrHandler_Acc(void) {
  unsigned char oldVal,newVal;
  /* Accumulator */
  oldVal=Accumulator;
  Accumulator=newVal=(((unsigned int)Accumulator)>>1) & 127;
  SetPSRCZN((oldVal & 1)>0, newVal==0,0);
} /* LSRInstrHandler_Acc */

INLINE static void ORAInstrHandler(int16 operand) {
  Accumulator=Accumulator | operand;
  SetPSRZN(Accumulator);
} /* ORAInstrHandler */

INLINE static void ROLInstrHandler(int16 address) {
  unsigned char oldVal,newVal;

  oldVal=TUBEREADMEM_FAST(address);
  newVal=((unsigned int)oldVal<<1) & 254;
  newVal+=GETCFLAG;
  TUBEWRITEMEM_FAST(address,newVal);
  SetPSRCZN((oldVal & 128)>0,newVal==0,newVal & 128);
} /* ROLInstrHandler */

INLINE static void ROLInstrHandler_Acc(void) {
  unsigned char oldVal,newVal;

  oldVal=Accumulator;
  newVal=((unsigned int)oldVal<<1) & 254;
  newVal+=GETCFLAG;
  Accumulator=newVal;
  SetPSRCZN((oldVal & 128)>0,newVal==0,newVal & 128);
} /* ROLInstrHandler_Acc */

INLINE static void RORInstrHandler(int16 address) {
  unsigned char oldVal,newVal;

  oldVal=TUBEREADMEM_FAST(address);
  newVal=((unsigned int)oldVal>>1) & 127;
  newVal+=GETCFLAG*128;
  TUBEWRITEMEM_FAST(address,newVal);
  SetPSRCZN(oldVal & 1,newVal==0,newVal & 128);
} /* RORInstrHandler */

INLINE static void RORInstrHandler_Acc(void) {
  unsigned char oldVal,newVal;

  oldVal=Accumulator;
  newVal=((unsigned int)oldVal>>1) & 127;
  newVal+=GETCFLAG*128;
  Accumulator=newVal;
  SetPSRCZN(oldVal & 1,newVal==0,newVal & 128);
} /* RORInstrHandler_Acc */

INLINE static void SBCInstrHandler(int16 operand) {
  /* NOTE! Not sure about C and V flags */
  if (!GETDFLAG) {
    int TmpResultV = (signed char)Accumulator - (signed char)operand -(1 - GETCFLAG);
    int TmpResultC = Accumulator - operand - (1 - GETCFLAG);
    Accumulator = TmpResultC & 255;
    SetPSR(FlagC | FlagZ | FlagV | FlagN, TmpResultC >= 0,
      Accumulator == 0, 0, 0, 0,
      ((Accumulator & 128) > 0) ^ ((TmpResultV & 256) != 0),
      Accumulator & 128);
  } else {
    int ohn = operand & 0xf0;
    int oln = operand & 0x0f;

    int ln = (Accumulator & 0xf) - oln - (1 - GETCFLAG);
    int TmpResult = Accumulator - operand - (1 - GETCFLAG);

    int TmpResultV = (signed char)Accumulator - (signed char)operand - (1 - GETCFLAG);
    int VFlag = ((TmpResultV < -128) || (TmpResultV > 127));

    int CFlag = (TmpResult & 256) == 0;

    if (TmpResult < 0) {
      TmpResult -= 0x60;
     }

    if (ln < 0) {
      TmpResult -= 0x06;
    }

    int NFlag = TmpResult & 128;
    Accumulator = TmpResult & 0xFF;
    int ZFlag = (Accumulator == 0);

    SetPSR(FlagC | FlagZ | FlagV | FlagN, CFlag, ZFlag, 0, 0, 0, VFlag, NFlag);
  }
} /* SBCInstrHandler */

INLINE static void STXInstrHandler(int16 address) {
  TUBEWRITEMEM_FAST(address, XReg);
} /* STXInstrHandler */

INLINE static void STYInstrHandler(int16 address) {
  TUBEWRITEMEM_FAST(address, YReg);
} /* STYInstrHandler */

/*-------------------------------------------------------------------------*/

// The RMB, SMB, BBR, and BBS instructions are specific to the 65C02,
// used in the Acorn 6502 co-processor. They are not implemented in the
// 65SC02, used in the Master 128.

static void ResetMemoryBit(int bit)
{
	const int EffectiveAddress = TubeRam[TubeProgramCounter++];

	TUBEWRITEMEM_DIRECT(EffectiveAddress, TubeRam[EffectiveAddress] & ~(1 << bit));
}

static void SetMemoryBit(int bit)
{
	const int EffectiveAddress = TubeRam[TubeProgramCounter++];

	TUBEWRITEMEM_DIRECT(EffectiveAddress, TubeRam[EffectiveAddress] | (1 << bit));
}

static void BranchOnBitReset(int bit)
{
	const int EffectiveAddress = TubeRam[TubeProgramCounter++];
	const int Offset = TubeRam[TubeProgramCounter++];

	if ((TubeRam[EffectiveAddress] & (1 << bit)) == 0) {
		TubeProgramCounter += Offset;
	}
}

static void BranchOnBitSet(int bit)
{
	const int EffectiveAddress = TubeRam[TubeProgramCounter++];
	const int Offset = TubeRam[TubeProgramCounter++];

	if (TubeRam[EffectiveAddress] & (1 << bit)) {
		TubeProgramCounter += Offset;
	}
}

/*-------------------------------------------------------------------------*/
/* Absolute  addressing mode handler                                       */
INLINE static int16 AbsAddrModeHandler_Data(void) {
  int FullAddress;

  /* Get the address from after the instruction */

  GETTWOBYTEFROMPC(FullAddress)

  /* And then read it */
  return(TUBEREADMEM_FAST(FullAddress));
} /* AbsAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Absolute  addressing mode handler                                       */
INLINE static int16 AbsAddrModeHandler_Address(void) {
  int FullAddress;

  /* Get the address from after the instruction */
  GETTWOBYTEFROMPC(FullAddress)

  /* And then read it */
  return(FullAddress);
} /* AbsAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Zero page addressing mode handler                                       */
INLINE static int16 ZeroPgAddrModeHandler_Address(void) {
  return(TubeRam[TubeProgramCounter++]);
} /* ZeroPgAddrModeHandler_Address */

/*-------------------------------------------------------------------------*/
/* Indexed with X preinc addressing mode handler                           */
INLINE static int16 IndXAddrModeHandler_Data(void) {
  unsigned char ZeroPageAddress;
  int EffectiveAddress;

  ZeroPageAddress=(TubeRam[TubeProgramCounter++]+XReg) & 255;

  EffectiveAddress=TubeRam[ZeroPageAddress] | (TubeRam[ZeroPageAddress+1]<<8);
  return(TUBEREADMEM_FAST(EffectiveAddress));
} /* IndXAddrModeHandler_Data */

/*-------------------------------------------------------------------------*/
/* Indexed with X preinc addressing mode handler                           */
INLINE static int16 IndXAddrModeHandler_Address(void) {
  unsigned char ZeroPageAddress;
  int EffectiveAddress;

  ZeroPageAddress=(TubeRam[TubeProgramCounter++]+XReg) & 255;

  EffectiveAddress=TubeRam[ZeroPageAddress] | (TubeRam[ZeroPageAddress+1]<<8);
  return(EffectiveAddress);
} /* IndXAddrModeHandler_Address */

/*-------------------------------------------------------------------------*/
/* Indexed with Y postinc addressing mode handler                          */
INLINE static int16 IndYAddrModeHandler_Data(void) {
  uint8_t ZPAddr=TubeRam[TubeProgramCounter++];
  uint16_t EffectiveAddress=TubeRam[ZPAddr]+YReg;
  if (EffectiveAddress>0xff) Carried();
  EffectiveAddress+=(TubeRam[(uint8_t)(ZPAddr+1)]<<8);

  return(TUBEREADMEM_FAST(EffectiveAddress));
} /* IndYAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Indexed with Y postinc addressing mode handler                          */
INLINE static int16 IndYAddrModeHandler_Address(void) {
  uint8_t ZPAddr=TubeRam[TubeProgramCounter++];
  uint16_t EffectiveAddress=TubeRam[ZPAddr]+YReg;
  if (EffectiveAddress>0xff) Carried();
  EffectiveAddress+=(TubeRam[(uint8_t)(ZPAddr+1)]<<8);

  return(EffectiveAddress);
} /* IndYAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Zero page wih X offset addressing mode handler                          */
INLINE static int16 ZeroPgXAddrModeHandler_Data(void) {
  int EffectiveAddress;
  EffectiveAddress=(TubeRam[TubeProgramCounter++]+XReg) & 255;
  return(TubeRam[EffectiveAddress]);
} /* ZeroPgXAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Zero page wih X offset addressing mode handler                          */
INLINE static int16 ZeroPgXAddrModeHandler_Address(void) {
  int EffectiveAddress;
  EffectiveAddress=(TubeRam[TubeProgramCounter++]+XReg) & 255;
  return(EffectiveAddress);
} /* ZeroPgXAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Absolute with X offset addressing mode handler                          */
INLINE static int16 AbsXAddrModeHandler_Data(void) {
  int EffectiveAddress;
  GETTWOBYTEFROMPC(EffectiveAddress);
  if ((EffectiveAddress & 0xff00)!=((EffectiveAddress+XReg) & 0xff00)) Carried();
  EffectiveAddress+=XReg;
  EffectiveAddress&=0xffff;

  return(TUBEREADMEM_FAST(EffectiveAddress));
} /* AbsXAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Absolute with X offset addressing mode handler                          */
INLINE static int16 AbsXAddrModeHandler_Address(void) {
  int EffectiveAddress;
  GETTWOBYTEFROMPC(EffectiveAddress)
  if ((EffectiveAddress & 0xff00)!=((EffectiveAddress+XReg) & 0xff00)) Carried();
  EffectiveAddress+=XReg;
  EffectiveAddress&=0xffff;

  return(EffectiveAddress);
} /* AbsXAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Absolute with Y offset addressing mode handler                          */
INLINE static int16 AbsYAddrModeHandler_Data(void) {
  int EffectiveAddress;
  GETTWOBYTEFROMPC(EffectiveAddress);
  if ((EffectiveAddress & 0xff00)!=((EffectiveAddress+YReg) & 0xff00)) Carried();
  EffectiveAddress+=YReg;
  EffectiveAddress&=0xffff;

  return(TUBEREADMEM_FAST(EffectiveAddress));
} /* AbsYAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Absolute with Y offset addressing mode handler                          */
INLINE static int16 AbsYAddrModeHandler_Address(void) {
  int EffectiveAddress;
  GETTWOBYTEFROMPC(EffectiveAddress)
  if ((EffectiveAddress & 0xff00)!=((EffectiveAddress+YReg) & 0xff00)) Carried();
  EffectiveAddress+=YReg;
  EffectiveAddress&=0xffff;

  return(EffectiveAddress);
} /* AbsYAddrModeHandler */

/*-------------------------------------------------------------------------*/

// Indirect addressing mode handler (for JMP indirect only)

INLINE static int16 IndAddrModeHandler_Address()
{
	int VectorLocation;
	GETTWOBYTEFROMPC(VectorLocation);

	// The 65C02 fixed the bug in the 6502 concerning this addressing mode
	// and VectorLocation == xxFF
	int EffectiveAddress = TUBEREADMEM_FAST(VectorLocation) |
	                       (TUBEREADMEM_FAST(VectorLocation + 1) << 8);

	return EffectiveAddress;
}

/*-------------------------------------------------------------------------*/
/* Zero page Indirect addressing mode handler                                        */
INLINE static int16 ZPIndAddrModeHandler_Address(void) {
  int VectorLocation;
  int EffectiveAddress;

  VectorLocation=TubeRam[TubeProgramCounter++];
  EffectiveAddress=TubeRam[VectorLocation]+(TubeRam[VectorLocation+1]<<8);

   // EffectiveAddress|=TUBEREADMEM_FAST(VectorLocation+1) << 8; }
  return(EffectiveAddress);
} /* ZPIndAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Zero page Indirect addressing mode handler                                        */
INLINE static int16 ZPIndAddrModeHandler_Data(void) {
  int VectorLocation;
  int EffectiveAddress;

  VectorLocation=TubeRam[TubeProgramCounter++];
  EffectiveAddress=TubeRam[VectorLocation]+(TubeRam[VectorLocation+1]<<8);

   // EffectiveAddress|=TUBEREADMEM_FAST(VectorLocation+1) << 8; }
  return(TubeRam[EffectiveAddress]);
} /* ZPIndAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Pre-indexed absolute Indirect addressing mode handler                                        */
INLINE static int16 IndAddrXModeHandler_Address(void) {
  /* For jump indirect only */
  int VectorLocation;
  int EffectiveAddress;

  GETTWOBYTEFROMPC(VectorLocation)
  EffectiveAddress=TUBEREADMEM_FAST(VectorLocation+XReg);
  EffectiveAddress|=TUBEREADMEM_FAST(VectorLocation+1+XReg) << 8;

   // EffectiveAddress|=TUBEREADMEM_FAST(VectorLocation+1) << 8; }
  return(EffectiveAddress);
} /* ZPIndAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Zero page with Y offset addressing mode handler                         */
INLINE static int16 ZeroPgYAddrModeHandler_Data(void) {
  int EffectiveAddress;
  EffectiveAddress=(TubeRam[TubeProgramCounter++]+YReg) & 255;
  return(TubeRam[EffectiveAddress]);
} /* ZeroPgYAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Zero page with Y offset addressing mode handler                         */
INLINE static int16 ZeroPgYAddrModeHandler_Address(void) {
  int EffectiveAddress;
  EffectiveAddress=(TubeRam[TubeProgramCounter++]+YReg) & 255;
  return(EffectiveAddress);
} /* ZeroPgYAddrModeHandler */

/*-------------------------------------------------------------------------*/
/* Reset processor */
static void Reset65C02()
{
  FILE *TubeRom;
  char TRN[256];
  char *TubeRomName=TRN;
  Accumulator=XReg=YReg=0; /* For consistancy of execution */
  StackReg=0xff; /* Initial value ? */
  PSR=FlagI; /* Interrupts off for starters */

  TubeintStatus=0;
  TubeNMIStatus=0;
  TubeNMILock = false;

  //The fun part, the tube OS is copied from ROM to tube RAM before the processor starts processing
  //This makes the OS "ROM" writable in effect, but must be restored on each reset.
  strcpy(TubeRomName,RomPath);
  strcat(TubeRomName,"beebfile/6502Tube.rom");
  TubeRom=fopen(TubeRomName,"rb");
  if (TubeRom!=NULL) {
	  fread(TubeRam+0xf800,1,2048,TubeRom);
	  fclose(TubeRom);
  }

  TubeProgramCounter=TubeReadMem(0xfffc) | (TubeReadMem(0xfffd)<<8);
  TotalTubeCycles=TotalCycles/2*3;
}

/* Reset Tube */
void ResetTube(void)
{
  memset(R1PHData,0,TubeBufferLength * 2);
  R1PHPtr=0;
  R1HStatus=TubeNotFull;
  R1HPData=0;
  R1PStatus=TubeNotFull;

  R2PHData=0;
  R2HStatus=TubeNotFull;
  R2HPData=0;
  R2PStatus=TubeNotFull;

  R3PHData[0]=0;
  R3PHData[1]=0;
  R3PHPtr=1; // To prevent NMI on reset
  R3HStatus=TubeDataAv | TubeNotFull;
  R3HPData[0]=0;
  R3HPData[1]=0;
  R3HPPtr=0;
  R3PStatus=TubeNotFull;

  R4PHData=0;
  R4HStatus=TubeNotFull;
  R4HPData=0;
  R4PStatus=TubeNotFull;

  TubeintStatus=0;
  TubeNMIStatus=0;
}

/* Initialise 6502core */
void Init65C02core(void) {
  Reset65C02();
  R1Status=0;
  ResetTube();
}

#include "via.h"

/*-------------------------------------------------------------------------*/
void DoTubeInterrupt(void) {
  PushWord(TubeProgramCounter);
  Push(PSR & ~FlagB);
  TubeProgramCounter=TubeReadMem(0xfffe) | (TubeReadMem(0xffff)<<8);
  SetPSR(FlagI,0,0,1,0,0,0,0);
  IRQCycles=7;
} /* DoInterrupt */

/*-------------------------------------------------------------------------*/
void DoTubeNMI(void) {
  /*cerr << "Doing NMI\n"; */
  TubeNMILock = true;
  PushWord(TubeProgramCounter);
  Push(PSR);
  TubeProgramCounter=TubeReadMem(0xfffa) | (TubeReadMem(0xfffb)<<8);
  SetPSR(FlagI,0,0,1,0,0,0,0); /* Normal interrupts should be disabled during NMI ? */
  IRQCycles=7;
} /* DoNMI */

/*-------------------------------------------------------------------------*/

// Execute one 6502 instruction, move program counter on

void Exec65C02Instruction() {
	static int tmpaddr;
	static unsigned char OldTubeNMIStatus;

	// Output debug info
	if (DebugEnabled) {
		DebugDisassembler(TubeProgramCounter, PreTPC, Accumulator, XReg, YReg, PSR, StackReg, false);
	}

	// For the Master, check Shadow Ram Presence
	// Note, this has to be done BEFORE reading an instruction due to Bit E and the PC
	int OldPC = TubeProgramCounter;
	PreTPC = TubeProgramCounter;

	// Read an instruction and post inc program counter
	CurrentInstruction = TubeRam[TubeProgramCounter++];
	// cout << "Fetch at " << hex << (TubeProgramCounter-1) << " giving 0x" << CurrentInstruction << dec << "\n";
	TubeCycles = TubeCyclesTable[CurrentInstruction];
	// Stats[CurrentInstruction]++;

	Branched = false;

	switch (CurrentInstruction) {
		case 0x00:
			// BRK
			BRKInstrHandler();
			break;
		case 0x01:
			// ORA (zp,X)
			ORAInstrHandler(IndXAddrModeHandler_Data());
			break;
		case 0x02:
		case 0x22:
		case 0x42:
		case 0x62:
		case 0x82:
		case 0xc2:
		case 0xe2:
			// NOP imm
			TubeProgramCounter++;
			break;
		case 0x03:
		case 0x13:
		case 0x23:
		case 0x33:
		case 0x43:
		case 0x53:
		case 0x63:
		case 0x73:
		case 0x83:
		case 0x93:
		case 0xa3:
		case 0xb3:
		case 0xc3:
		case 0xd3:
		case 0xe3:
		case 0xf3:
			// NOP
			break;
		case 0x04:
			// TSB zp
			TSBInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0x05:
			// ORA zp
			ORAInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0x06:
			// ASL zp
			ASLInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0x07:
			// RMB0
			ResetMemoryBit(0);
			break;
		case 0x08:
			// PHP
			Push(PSR | 48);
			break;
		case 0x09:
			// ORA imm
			ORAInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0x0a:
			// ASL A
			ASLInstrHandler_Acc();
			break;
		case 0x0b:
		case 0x1b:
		case 0x2b:
		case 0x3b:
		case 0x4b:
		case 0x5b:
		case 0x6b:
		case 0x7b:
		case 0x8b:
		case 0x9b:
		case 0xab:
		case 0xbb:
		case 0xcb:
		case 0xdb:
		case 0xeb:
		case 0xfb:
			// NOP
			break;
		case 0x0c:
			// TSB abs
			TSBInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x0d:
			// ORA abs
			ORAInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0x0e:
			// ASL abs
			ASLInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x0f:
			// BBR0
			BranchOnBitReset(0);
			break;
		case 0x10:
			// BPL rel
			BPLInstrHandler();
			break;
		case 0x11:
			// ORA (zp),Y
			ORAInstrHandler(IndYAddrModeHandler_Data());
			break;
		case 0x12:
			// ORA (zp)
			ORAInstrHandler(ZPIndAddrModeHandler_Data());
			break;
		case 0x14:
			// TRB zp
			TRBInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0x15:
			// ORA zp,X
			ORAInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0x16:
			// ASL zp,X
			ASLInstrHandler(ZeroPgXAddrModeHandler_Address());
			break;
		case 0x17:
			// RMB1
			ResetMemoryBit(1);
			break;
		case 0x18:
			// CLC
			PSR &= 255 - FlagC;
			break;
		case 0x19:
			// ORA abs,Y
			ORAInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0x1a:
			// INC A
			INAInstrHandler();
			break;
		case 0x1c:
			// TRB abs
			TRBInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x1d:
			// ORA abs,X
			ORAInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0x1e:
			// ASL abs,X
			ASLInstrHandler(AbsXAddrModeHandler_Address());
			break;
		case 0x1f:
			// BBR1
			BranchOnBitReset(1);
			break;
		case 0x20:
			// JSR abs
			JSRInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x21:
			// AND (zp,X)
			ANDInstrHandler(IndXAddrModeHandler_Data());
			break;
		case 0x24:
			// BIT zp
			BITInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0x25:
			// AND zp
			ANDInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0x26:
			// ROL zp
			ROLInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0x27:
			// RMB2
			ResetMemoryBit(2);
			break;
		case 0x28:
			// PLP
			PSR = Pop();
			break;
		case 0x29:
			// AND imm
			ANDInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0x2a:
			// ROL A
			ROLInstrHandler_Acc();
			break;
		case 0x2c:
			// BIT abs
			BITInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0x2d:
			// AND abs
			ANDInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0x2e:
			// ROL abs
			ROLInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x2f:
			// BBR2
			BranchOnBitReset(2);
			break;
		case 0x30:
			// BMI rel
			BMIInstrHandler();
			break;
		case 0x31:
			// AND (zp),Y
			ANDInstrHandler(IndYAddrModeHandler_Data());
			break;
		case 0x32:
			// AND (zp)
			ANDInstrHandler(ZPIndAddrModeHandler_Data());
			break;
		case 0x34:
			// BIT abs,x
			BITInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0x35:
			// AND zp,X
			ANDInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0x36:
			// ROL zp,X
			ROLInstrHandler(ZeroPgXAddrModeHandler_Address());
			break;
		case 0x37:
			// RMB3
			ResetMemoryBit(3);
			break;
		case 0x38:
			// SEC
			PSR |= FlagC;
			break;
		case 0x39:
			ANDInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0x3a:
			// DEC A
			DEAInstrHandler();
			break;
		case 0x3c:
			// BIT abs,x
			BITInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0x3d:
			// AND abs,X
			ANDInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0x3e:
			// ROL abs,X
			ROLInstrHandler(AbsXAddrModeHandler_Address());
			break;
		case 0x3f:
			// BBR3
			BranchOnBitReset(3);
			break;
		case 0x40:
			// RTI
			PSR = Pop();
			TubeProgramCounter = PopWord();
			TubeNMILock = false;
			break;
		case 0x41:
			// EOR (zp,X)
			EORInstrHandler(IndXAddrModeHandler_Data());
			break;
		case 0x44:
			// NOP zp
			TubeProgramCounter += 1;
			break;
		case 0x45:
			// EOR zp
			EORInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0x46:
			// LSR zp
			LSRInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0x47:
			// RMB4
			ResetMemoryBit(4);
			break;
		case 0x48:
			// PHA
			Push(Accumulator);
			break;
		case 0x49:
			// EOR imm
			EORInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0x4a:
			// LSR A
			LSRInstrHandler_Acc();
			break;
		case 0x4c:
			// JMP
			TubeProgramCounter = AbsAddrModeHandler_Address();
			break;
		case 0x4d:
			// EOR abs
			EORInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0x4e:
			// LSR abs
			LSRInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x4f:
			// BBR4
			BranchOnBitReset(4);
			break;
		case 0x50:
			// BVC rel
			BVCInstrHandler();
			break;
		case 0x51:
			// EOR (zp),Y
			EORInstrHandler(IndYAddrModeHandler_Data());
			break;
		case 0x52:
			// EOR (zp)
			EORInstrHandler(ZPIndAddrModeHandler_Data());
			break;
		case 0x54:
			// NOP zp,X
			TubeProgramCounter += 1;
			break;
		case 0x55:
			// EOR zp,X
			EORInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0x56:
			// LSR zp,X
			LSRInstrHandler(ZeroPgXAddrModeHandler_Address());
			break;
		case 0x57:
			// RMB5
			ResetMemoryBit(5);
			break;
		case 0x58:
			// CLI
			PSR &= 255 - FlagI;
			break;
		case 0x59:
			// EOR abs,Y
			EORInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0x5a:
			// PHY
			Push(YReg);
			break;
		case 0x5c:
			// NOP abs
			TubeProgramCounter += 2;
			break;
		case 0x5d:
			// EOR abs,X
			EORInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0x5e:
			// LSR abs,X
			LSRInstrHandler(AbsXAddrModeHandler_Address());
			break;
		case 0x5f:
			// BBR5
			BranchOnBitReset(5);
			break;
		case 0x60:
			// RTS
			TubeProgramCounter = PopWord() + 1;
			break;
		case 0x61:
			// ADC (zp,X)
			ADCInstrHandler(IndXAddrModeHandler_Data());
			break;
		case 0x64:
			// STZ zp
			TUBEWRITEMEM_DIRECT(ZeroPgAddrModeHandler_Address(), 0);
			break;
		case 0x65:
			// ADC zp
			ADCInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0x66:
			// ROR zp
			RORInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0x67:
			// RMB6
			ResetMemoryBit(6);
			break;
		case 0x68:
			// PLA
			Accumulator = Pop();
			SetPSRZN(Accumulator);
			break;
		case 0x69:
			// ADC imm
			ADCInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0x6a:
			// ROR A
			RORInstrHandler_Acc();
			break;
		case 0x6c:
			// JMP (abs)
			TubeProgramCounter = IndAddrModeHandler_Address();
			break;
		case 0x6d:
			// ADC abs
			ADCInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0x6e:
			// ROR abs
			RORInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x6f:
			// BBR6
			BranchOnBitReset(6);
			break;
		case 0x70:
			// BVS rel
			BVSInstrHandler();
			break;
		case 0x71:
			// ADC (zp),Y
			ADCInstrHandler(IndYAddrModeHandler_Data());
			break;
		case 0x72:
			// ADC (zp)
			ADCInstrHandler(ZPIndAddrModeHandler_Data());
			break;
		case 0x74:
			// STZ zp,x
			TUBEFASTWRITE(ZeroPgXAddrModeHandler_Address(), 0);
			break;
		case 0x75:
			// ADC zp,X
			ADCInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0x76:
			// ROR zp,X
			RORInstrHandler(ZeroPgXAddrModeHandler_Address());
			break;
		case 0x77:
			// RMB7
			ResetMemoryBit(7);
			break;
		case 0x78:
			// SEI
			PSR |= FlagI;
			break;
		case 0x79:
			// ADC abs,Y
			ADCInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0x7a:
			// PLY
			YReg = Pop();
			SetPSRZN(YReg);
			break;
		case 0x7c:
			// JMP abs,X
			TubeProgramCounter = IndAddrXModeHandler_Address();
			break;
		case 0x7d:
			// ADC abs,X
			ADCInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0x7e:
			// ROR abs,X
			RORInstrHandler(AbsXAddrModeHandler_Address());
			break;
		case 0x7f:
			// BBR7
			BranchOnBitReset(7);
			break;
		case 0x80:
			// BRA rel
			BRAInstrHandler();
			break;
		case 0x81:
			// STA (zp,X)
			TUBEFASTWRITE(IndXAddrModeHandler_Address(), Accumulator);
			break;
		case 0x84:
			// STY zp
			TUBEWRITEMEM_DIRECT(ZeroPgAddrModeHandler_Address(), YReg);
			break;
		case 0x85:
			// STA zp
			TUBEWRITEMEM_DIRECT(ZeroPgAddrModeHandler_Address(), Accumulator);
			break;
		case 0x86:
			// STX zp
			TUBEWRITEMEM_DIRECT(ZeroPgAddrModeHandler_Address(), XReg);
			break;
		case 0x87:
			// SMB0
			SetMemoryBit(0);
			break;
		case 0x88:
			// DEY
			YReg = (YReg - 1) & 0xff;
			SetPSRZN(YReg);
			break;
		case 0x89:
			// BIT imm
			BITImmedInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0x8a:
			// TXA
			Accumulator = XReg;
			SetPSRZN(Accumulator);
			break;
		case 0x8c:
			// STY abs
			STYInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x8d:
			// STA abs
			TUBEFASTWRITE(AbsAddrModeHandler_Address(), Accumulator);
			break;
		case 0x8e:
			// STX abs
			STXInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0x8f:
			// BBS0
			BranchOnBitSet(0);
			break;
		case 0x90:
			// BCC rel
			BCCInstrHandler();
			break;
		case 0x91:
			// STA (zp),Y
			TUBEFASTWRITE(IndYAddrModeHandler_Address(), Accumulator);
			break;
		case 0x92:
			// STA (zp)
			TUBEFASTWRITE(ZPIndAddrModeHandler_Address(), Accumulator);
			break;
		case 0x94:
			// STY zp,X
			STYInstrHandler(ZeroPgXAddrModeHandler_Address());
			break;
		case 0x95:
			// STA zp,X
			TUBEFASTWRITE(ZeroPgXAddrModeHandler_Address(), Accumulator);
			break;
		case 0x96:
			// STX zp,X
			STXInstrHandler(ZeroPgYAddrModeHandler_Address());
			break;
		case 0x97:
			// SMB1
			SetMemoryBit(1);
			break;
		case 0x98:
			// TYA
			Accumulator = YReg;
			SetPSRZN(Accumulator);
			break;
		case 0x99:
			// STA abs,Y
			TUBEFASTWRITE(AbsYAddrModeHandler_Address(), Accumulator);
			break;
		case 0x9a:
			// TXS
			StackReg = XReg;
			break;
		case 0x9c:
			// STZ abs
			TUBEFASTWRITE(AbsAddrModeHandler_Address(), 0);
			break;
		case 0x9d:
			// STA abs,X
			TUBEFASTWRITE(AbsXAddrModeHandler_Address(), Accumulator);
			break;
		case 0x9e:
			// STZ abs,x
			TUBEFASTWRITE(AbsXAddrModeHandler_Address(), 0);
			break;
		case 0x9f:
			// BBS1
			BranchOnBitSet(1);
			break;
		case 0xa0:
			// LDY imm
			LDYInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0xa1:
			// LDA (zp,X)
			LDAInstrHandler(IndXAddrModeHandler_Data());
			break;
		case 0xa2:
			// LDX imm
			LDXInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0xa4:
			// LDY zp
			LDYInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0xa5:
			// LDA zp
			LDAInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0xa6:
			// LDX zp
			LDXInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0xa7:
			// SMB2
			SetMemoryBit(2);
			break;
		case 0xa8:
			// TAY
			YReg = Accumulator;
			SetPSRZN(Accumulator);
			break;
		case 0xa9:
			// LDA imm
			LDAInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0xaa:
			// TXA
			XReg = Accumulator;
			SetPSRZN(Accumulator);
			break;
		case 0xaf:
			// BBS2
			BranchOnBitSet(2);
			break;
		case 0xac:
			// LDY abs
			LDYInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0xad:
			// LDA abs
			LDAInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0xae:
			// LDX abs
			LDXInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0xb0:
			// BCS rel
			BCSInstrHandler();
			break;
		case 0xb1:
			// LDA (zp),Y
			LDAInstrHandler(IndYAddrModeHandler_Data());
			break;
		case 0xb2:
			// LDA (zp)
			LDAInstrHandler(ZPIndAddrModeHandler_Data());
			break;
		case 0xb4:
			// LDY zp,X
			LDYInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0xb5:
			// LDA zp,X
			LDAInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0xb6:
			// LDX zp,X
			LDXInstrHandler(ZeroPgYAddrModeHandler_Data());
			break;
		case 0xb7:
			// SMB3
			SetMemoryBit(3);
			break;
		case 0xb8:
			// CLV
			PSR &= 255 - FlagV;
			break;
		case 0xb9:
			// LDA abs,Y
			LDAInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0xba:
			// TSX
			XReg = StackReg;
			SetPSRZN(XReg);
			break;
		case 0xbc:
			// LDY abs,X
			LDYInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0xbd:
			// LDA abs,X
			LDAInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0xbe:
			// LDX abs,X
			LDXInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0xbf:
			// BBS3
			BranchOnBitSet(3);
			break;
		case 0xc0:
			// CPY imm
			CPYInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0xc1:
			// CMP (zp,X)
			CMPInstrHandler(IndXAddrModeHandler_Data());
			break;
		case 0xc4:
			// CPY zp
			CPYInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0xc5:
			// CMP zp
			CMPInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0xc6:
			// DEC zp
			DECInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0xc7:
			// SMB4
			SetMemoryBit(4);
			break;
		case 0xc8:
			// INY
			YReg += 1;
			YReg &= 255;
			SetPSRZN(YReg);
			break;
		case 0xc9:
			// CMP imm
			CMPInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0xca:
			// DEX
			DEXInstrHandler();
			break;
		case 0xcc:
			// CPY abs
			CPYInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0xcd:
			// CMP abs
			CMPInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0xce:
			// DEC abs
			DECInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0xcf:
			// BBS4
			BranchOnBitSet(4);
			break;
		case 0xd0:
			// BNE rel
			BNEInstrHandler();
			break;
		case 0xd1:
			CMPInstrHandler(IndYAddrModeHandler_Data());
			break;
		case 0xd2:
			// CMP (zp)
			CMPInstrHandler(ZPIndAddrModeHandler_Data());
			break;
		case 0xd4:
			// NOP zp,X
			TubeProgramCounter += 1;
			break;
		case 0xd5:
			// CMP zp,X
			CMPInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0xd6:
			// DEC zp,X
			DECInstrHandler(ZeroPgXAddrModeHandler_Address());
			break;
		case 0xd7:
			// SMB5
			SetMemoryBit(5);
			break;
		case 0xd8:
			// CLD
			PSR &= 255 - FlagD;
			break;
		case 0xd9:
			// CMP abs,Y
			CMPInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0xda:
			// PHX
			Push(XReg);
			break;
		case 0xdc:
			// NOP abs
			TubeProgramCounter += 2;
			break;
		case 0xdd:
			// CMP abs,X
			CMPInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0xde:
			// DEC abs,X
			DECInstrHandler(AbsXAddrModeHandler_Address());
			break;
		case 0xdf:
			// BBS5
			BranchOnBitSet(5);
			break;
		case 0xe0:
			// CPX imm
			CPXInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0xe1:
			// SBC (zp,X)
			SBCInstrHandler(IndXAddrModeHandler_Data());
			break;
		case 0xe4:
			// CPX zp
			CPXInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0xe5:
			// SBC zp
			SBCInstrHandler(TubeRam[TubeRam[TubeProgramCounter++]]);
			break;
		case 0xe6:
			// INC zp
			INCInstrHandler(ZeroPgAddrModeHandler_Address());
			break;
		case 0xe7:
			// SMB6
			SetMemoryBit(6);
			break;
		case 0xe8:
			// INX
			INXInstrHandler();
			break;
		case 0xe9:
			// SBC imm
			SBCInstrHandler(TubeRam[TubeProgramCounter++]);
			break;
		case 0xea:
			// NOP
			break;
		case 0xec:
			// CPX abs
			CPXInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0xed:
			// SBC abs
			SBCInstrHandler(AbsAddrModeHandler_Data());
			break;
		case 0xee:
			// INC abs
			INCInstrHandler(AbsAddrModeHandler_Address());
			break;
		case 0xef:
			// BBS6
			BranchOnBitSet(6);
			break;
		case 0xf0:
			// BEQ rel
			BEQInstrHandler();
			break;
		case 0xf1:
			// SBC (zp),Y
			SBCInstrHandler(IndYAddrModeHandler_Data());
			break;
		case 0xf2:
			// SBC (zp)
			SBCInstrHandler(ZPIndAddrModeHandler_Data());
			break;
		case 0xf4:
			// NOP zp,X
			TubeProgramCounter += 1;
			break;
		case 0xf5:
			// SBC zp,X
			SBCInstrHandler(ZeroPgXAddrModeHandler_Data());
			break;
		case 0xf6:
			// INC zp,X
			INCInstrHandler(ZeroPgXAddrModeHandler_Address());
			break;
		case 0xf7:
			// SMB7
			SetMemoryBit(7);
			break;
		case 0xf8:
			// SED
			PSR |= FlagD;
			break;
		case 0xf9:
			// SBC abs,Y
			SBCInstrHandler(AbsYAddrModeHandler_Data());
			break;
		case 0xfa:
			// PLX
			XReg = Pop();
			SetPSRZN(XReg);
			break;
		case 0xfc:
			// NOP abs
			TubeProgramCounter += 2;
			break;
		case 0xfd:
			// SBC abs,X
			SBCInstrHandler(AbsXAddrModeHandler_Data());
			break;
		case 0xfe:
			// INC abs,X
			INCInstrHandler(AbsXAddrModeHandler_Address());
			break;
		case 0xff:
			// BBS7
			BranchOnBitSet(7);
			break;
	}

	// This block corrects the cycle count for the branch instructions
	if ((CurrentInstruction == 0x10) ||
	    (CurrentInstruction == 0x30) ||
	    (CurrentInstruction == 0x50) ||
	    (CurrentInstruction == 0x70) ||
	    (CurrentInstruction == 0x80) ||
	    (CurrentInstruction == 0x90) ||
	    (CurrentInstruction == 0xb0) ||
	    (CurrentInstruction == 0xd0) ||
	    (CurrentInstruction == 0xf0))
	{
		if (Branched)
		{
			TubeCycles++;
			if ((TubeProgramCounter & 0xff00) != ((OldPC+2) & 0xff00)) {
				TubeCycles++;
			}
		}
	}

	TubeCycles += IRQCycles;
	IRQCycles = 0; // IRQ Timing
	// End of cycle correction

	if (TubeintStatus && !GETIFLAG) {
		DoTubeInterrupt();
	}

	TotalTubeCycles += TubeCycles;

	if (TubeNMIStatus && !OldTubeNMIStatus) {
		DoTubeNMI();
	}

	OldTubeNMIStatus = TubeNMIStatus;
}

/*-------------------------------------------------------------------------*/
void WrapTubeCycles(void) {
	TotalTubeCycles -= CycleCountWrap/2*3;
}

void SyncTubeProcessor(void) {
	// This proc syncronises the two processors on a cycle based timing.
	// Second pro runs at 3MHz
	while (TotalTubeCycles<(TotalCycles/2*3)) {
		Exec65C02Instruction();
	}
}

/*-------------------------------------------------------------------------*/
void DebugTubeState(void)
{
	DebugDisplayInfo("");

	DebugDisplayInfoF("HostTube: R1=%02X R2=%02X R3=%02X R4=%02X R1n=%02X R3n=%02X",
		(int)R1HStatus | R1Status,
		(int)R2HStatus,
		(int)R3HStatus,
		(int)R4HStatus,
		(int)R1PHPtr,
		(int)R3PHPtr);

	DebugDisplayInfoF("ParaTube: R1=%02X R2=%02X R3=%02X R4=%02X R3n=%02X",
		(int)R1PStatus | R1Status,
		(int)R2PStatus,
		(int)R3PStatus,
		(int)R4PStatus,
		(int)R3HPPtr);
}

/*-------------------------------------------------------------------------*/
void SaveTubeUEF(FILE *SUEF) {
	fput16(0x0470,SUEF);
	fput32(45,SUEF);
	fputc(R1Status,SUEF);
	fwrite(R1PHData,1,TubeBufferLength,SUEF);
	fputc(R1PHPtr,SUEF);
	fputc(R1HStatus,SUEF);
	fputc(R1HPData,SUEF);
	fputc(R1PStatus,SUEF);
	fputc(R2PHData,SUEF);
	fputc(R2HStatus,SUEF);
	fputc(R2HPData,SUEF);
	fputc(R2PStatus,SUEF);
	fputc(R3PHData[0],SUEF);
	fputc(R3PHData[1],SUEF);
	fputc(R3PHPtr,SUEF);
	fputc(R3HStatus,SUEF);
	fputc(R3HPData[0],SUEF);
	fputc(R3HPData[1],SUEF);
	fputc(R3HPPtr,SUEF);
	fputc(R3PStatus,SUEF);
	fputc(R4PHData,SUEF);
	fputc(R4HStatus,SUEF);
	fputc(R4HPData,SUEF);
	fputc(R4PStatus,SUEF);
}

void Save65C02UEF(FILE *SUEF) {
	fput16(0x0471,SUEF);
	fput32(16,SUEF);
	fput16(TubeProgramCounter,SUEF);
	fputc(Accumulator,SUEF);
	fputc(XReg,SUEF);
	fputc(YReg,SUEF);
	fputc(StackReg,SUEF);
	fputc(PSR,SUEF);
	fput32(TotalTubeCycles,SUEF);
	fputc(TubeintStatus,SUEF);
	fputc(TubeNMIStatus,SUEF);
	fputc(TubeNMILock,SUEF);
	fput16(0,SUEF);
}

void Save65C02MemUEF(FILE *SUEF) {
	fput16(0x0472,SUEF);
	fput32(65536,SUEF);
	fwrite(TubeRam,1,65536,SUEF);
}

void LoadTubeUEF(FILE *SUEF) {
	R1Status = fget8(SUEF);
	fread(R1PHData,1,TubeBufferLength,SUEF);
	R1PHPtr = fget8(SUEF);
	R1HStatus = fget8(SUEF);
	R1HPData = fget8(SUEF);
	R1PStatus = fget8(SUEF);
	R2PHData = fget8(SUEF);
	R2HStatus = fget8(SUEF);
	R2HPData = fget8(SUEF);
	R2PStatus = fget8(SUEF);
	R3PHData[0] = fget8(SUEF);
	R3PHData[1] = fget8(SUEF);
	R3PHPtr = fget8(SUEF);
	R3HStatus = fget8(SUEF);
	R3HPData[0] = fget8(SUEF);
	R3HPData[1] = fget8(SUEF);
	R3HPPtr = fget8(SUEF);
	R3PStatus = fget8(SUEF);
	R4PHData = fget8(SUEF);
	R4HStatus = fget8(SUEF);
	R4HPData = fget8(SUEF);
	R4PStatus = fget8(SUEF);
}

void Load65C02UEF(FILE *SUEF) {
	TubeProgramCounter = fget16(SUEF);
	Accumulator = fget8(SUEF);
	XReg = fget8(SUEF);
	YReg = fget8(SUEF);
	StackReg = fget8(SUEF);
	PSR = fget8(SUEF);
	// TotalTubeCycles = fget32(SUEF);
	int Dlong = fget32(SUEF);
	TubeintStatus = fget8(SUEF);
	TubeNMIStatus = fget8(SUEF);
	TubeNMILock = fgetbool(SUEF);
}

void Load65C02MemUEF(FILE *SUEF) {
	fread(TubeRam,1,65536,SUEF);
}
